"""
"""

__all__ = [
    "get_channel_configs",
    "VectorBus",
    "VectorBusParams",
    "VectorCanFdParams",
    "VectorCanParams",
    "VectorChannelConfig",
    "VectorError",
    "VectorInitializationError",
    "VectorOperationError",
]

from .canlib import (
    VectorBus,
    VectorBusParams,
    VectorCanFdParams,
    VectorCanParams,
    VectorChannelConfig,
    get_channel_configs,
)
from .exceptions import VectorError, VectorInitializationError, VectorOperationError



#!/usr/bin/env python

import shutil
import tempfile
import unittest
from tempfile import NamedTemporaryFile

import can


class LoadFileConfigTest(unittest.TestCase):
    configuration = {
        "default": {"interface": "virtual", "channel": "0"},
        "one": {"interface": "kvaser", "channel": "1"},
        "two": {"channel": "2"},
        "three": {"extra": "extra value"},
    }

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def _gen_configration_file(self, sections):
        with NamedTemporaryFile(
            mode="w", dir=self.test_dir, delete=False
        ) as tmp_config_file:
            content = []
            for section in sections:
                content.append(f"[{section}]")
                for k, v in self.configuration[section].items():
                    content.append(f"{k} = {v}")
            tmp_config_file.write("\n".join(content))
            return tmp_config_file.name

    def test_config_file_with_default(self):
        tmp_config = self._gen_configration_file(["default"])
        config = can.util.load_file_config(path=tmp_config)
        self.assertEqual(config, self.configuration["default"])

    def test_config_file_with_default_and_section(self):
        tmp_config = self._gen_configration_file(["default", "one"])

        config = can.util.load_file_config(path=tmp_config)
        self.assertEqual(config, self.configuration["default"])

        config.update(can.util.load_file_config(path=tmp_config, section="one"))
        self.assertEqual(config, self.configuration["one"])

    def test_config_file_with_section_only(self):
        tmp_config = self._gen_configration_file(["one"])
        config = can.util.load_file_config(path=tmp_config)
        config.update(can.util.load_file_config(path=tmp_config, section="one"))
        self.assertEqual(config, self.configuration["one"])

    def test_config_file_with_section_and_key_in_default(self):
        expected = self.configuration["default"].copy()
        expected.update(self.configuration["two"])

        tmp_config = self._gen_configration_file(["default", "two"])
        config = can.util.load_file_config(path=tmp_config)
        config.update(can.util.load_file_config(path=tmp_config, section="two"))
        self.assertEqual(config, expected)

    def test_config_file_with_section_missing_interface(self):
        expected = self.configuration["two"].copy()
        tmp_config = self._gen_configration_file(["two"])
        config = can.util.load_file_config(path=tmp_config)
        config.update(can.util.load_file_config(path=tmp_config, section="two"))
        self.assertEqual(config, expected)

    def test_config_file_extra(self):
        expected = self.configuration["default"].copy()
        expected.update(self.configuration["three"])

        tmp_config = self._gen_configration_file(["default", "three"])
        config = can.util.load_file_config(path=tmp_config)
        config.update(can.util.load_file_config(path=tmp_config, section="three"))
        self.assertEqual(config, expected)

    def test_config_file_with_non_existing_section(self):
        expected = self.configuration["default"].copy()

        tmp_config = self._gen_configration_file(["default", "one", "two", "three"])
        config = can.util.load_file_config(path=tmp_config)
        config.update(can.util.load_file_config(path=tmp_config, section="zero"))
        self.assertEqual(config, expected)


if __name__ == "__main__":
    unittest.main()



"""
Intrepid Control Systems (ICS) neoVI interface module.

python-ics is a Python wrapper around the API provided by Intrepid Control
Systems for communicating with their neoVI range of devices.

Implementation references:
* https://github.com/intrepidcs/python_ics
"""

import logging
import os
import tempfile
from collections import Counter, defaultdict, deque
from itertools import cycle
from threading import Event
from warnings import warn

from can import BusABC, Message

from ...exceptions import (
    CanError,
    CanInitializationError,
    CanOperationError,
    CanTimeoutError,
)

logger = logging.getLogger(__name__)

try:
    import ics
except ImportError as ie:
    logger.warning(
        "You won't be able to use the ICS neoVI can backend without the "
        "python-ics module installed!: %s",
        ie,
    )
    ics = None


try:
    from filelock import FileLock
except ImportError as ie:
    logger.warning(
        "Using ICS neoVI can backend without the "
        "filelock module installed may cause some issues!: %s",
        ie,
    )

    class FileLock:
        """Dummy file lock that does not actually do anything"""

        def __init__(self, lock_file, timeout=-1):
            self._lock_file = lock_file
            self.timeout = timeout

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            return None


# Use inter-process mutex to prevent concurrent device open.
# When neoVI server is enabled, there is an issue with concurrent device open.
open_lock = FileLock(os.path.join(tempfile.gettempdir(), "neovi.lock"))
description_id = cycle(range(1, 0x8000))


class ICSApiError(CanError):
    """
    Indicates an error with the ICS API.
    """

    # A critical error which affects operation or accuracy.
    ICS_SPY_ERR_CRITICAL = 0x10
    # An error which is not understood.
    ICS_SPY_ERR_QUESTION = 0x20
    # An important error which may be critical depending on the application
    ICS_SPY_ERR_EXCLAMATION = 0x30
    # An error which probably does not need attention.
    ICS_SPY_ERR_INFORMATION = 0x40

    def __init__(
        self,
        error_code: int,
        description_short: str,
        description_long: str,
        severity: int,
        restart_needed: int,
    ):
        super().__init__(f"{description_short}. {description_long}", error_code)
        self.description_short = description_short
        self.description_long = description_long
        self.severity = severity
        self.restart_needed = restart_needed == 1

    def __reduce__(self):
        return type(self), (
            self.error_code,
            self.description_short,
            self.description_long,
            self.severity,
            self.restart_needed,
        )

    @property
    def error_number(self) -> int:
        """Deprecated. Renamed to :attr:`can.CanError.error_code`."""
        warn(
            "ICSApiError::error_number has been renamed to error_code defined by CanError",
            DeprecationWarning,
        )
        return self.error_code

    @property
    def is_critical(self) -> bool:
        return self.severity == self.ICS_SPY_ERR_CRITICAL


class ICSInitializationError(ICSApiError, CanInitializationError):
    pass


class ICSOperationError(ICSApiError, CanOperationError):
    pass


class NeoViBus(BusABC):
    """
    The CAN Bus implemented for the python_ics interface
    https://github.com/intrepidcs/python_ics
    """

    def __init__(self, channel, can_filters=None, **kwargs):
        """
        :param channel:
            The channel ids to create this bus with.
            Can also be a single integer, netid name or a comma separated
            string.
        :type channel: int or str or list(int) or list(str)
        :param list can_filters:
            See :meth:`can.BusABC.set_filters` for details.
        :param bool receive_own_messages:
            If transmitted messages should also be received by this bus.
        :param bool use_system_timestamp:
            Use system timestamp for can messages instead of the hardware time
            stamp
        :param str serial:
            Serial to connect (optional, will use the first found if not
            supplied)
        :param int bitrate:
            Channel bitrate in bit/s. (optional, will enable the auto bitrate
            feature if not supplied)
        :param bool fd:
            If CAN-FD frames should be supported.
        :param int data_bitrate:
            Which bitrate to use for data phase in CAN FD.
            Defaults to arbitration bitrate.
        :param override_library_name:
            Absolute path or relative path to the library including filename.

        :raise ImportError:
            If *python-ics* is not available
        :raise CanInitializationError:
            If the bus could not be set up.
            May or may not be a :class:`~ICSInitializationError`.
        """
        if ics is None:
            raise ImportError("Please install python-ics")

        super().__init__(channel=channel, can_filters=can_filters, **kwargs)

        logger.info(f"CAN Filters: {can_filters}")
        logger.info(f"Got configuration of: {kwargs}")

        if "override_library_name" in kwargs:
            ics.override_library_name(kwargs.get("override_library_name"))

        if isinstance(channel, (list, tuple)):
            self.channels = channel
        elif isinstance(channel, int):
            self.channels = [channel]
        else:
            # Assume comma separated string of channels
            self.channels = [ch.strip() for ch in channel.split(",")]
        self.channels = [NeoViBus.channel_to_netid(ch) for ch in self.channels]

        type_filter = kwargs.get("type_filter")
        serial = kwargs.get("serial")
        self.dev = self._find_device(type_filter, serial)

        with open_lock:
            ics.open_device(self.dev)

        try:
            if "bitrate" in kwargs:
                for channel in self.channels:
                    ics.set_bit_rate(self.dev, kwargs.get("bitrate"), channel)

            if kwargs.get("fd", False):
                if "data_bitrate" in kwargs:
                    for channel in self.channels:
                        ics.set_fd_bit_rate(
                            self.dev, kwargs.get("data_bitrate"), channel
                        )
        except ics.RuntimeError as re:
            logger.error(re)
            err = ICSInitializationError(*ics.get_last_api_error(self.dev))
            try:
                self.shutdown()
            finally:
                raise err

        self._use_system_timestamp = bool(kwargs.get("use_system_timestamp", False))
        self._receive_own_messages = kwargs.get("receive_own_messages", True)

        self.channel_info = "{} {} CH:{}".format(
            self.dev.Name,
            self.get_serial_number(self.dev),
            self.channels,
        )
        logger.info(f"Using device: {self.channel_info}")

        self.rx_buffer = deque()
        self.message_receipts = defaultdict(Event)

    @staticmethod
    def channel_to_netid(channel_name_or_id):
        try:
            channel = int(channel_name_or_id)
        except ValueError:
            netid = f"NETID_{channel_name_or_id.upper()}"
            if hasattr(ics, netid):
                channel = getattr(ics, netid)
            else:
                raise ValueError(
                    "channel must be an integer or a valid ICS channel name"
                ) from None
        return channel

    @staticmethod
    def get_serial_number(device):
        """Decode (if needed) and return the ICS device serial string

        :param device: ics device
        :return: ics device serial string
        :rtype: str
        """
        if int("AA0000", 36) < device.SerialNumber < int("ZZZZZZ", 36):
            return ics.base36enc(device.SerialNumber)
        else:
            return str(device.SerialNumber)

    def shutdown(self):
        super().shutdown()
        ics.close_device(self.dev)

    @staticmethod
    def _detect_available_configs():
        """Detect all configurations/channels that this interface could
        currently connect with.

        :rtype: Iterator[dict]
        :return: an iterable of dicts, each being a configuration suitable
                 for usage in the interface's bus constructor.
        """
        if ics is None:
            return []

        try:
            devices = ics.find_devices()
        except Exception as e:
            logger.debug("Failed to detect configs: %s", e)
            return []

        # TODO: add the channel(s)
        return [
            {"interface": "neovi", "serial": NeoViBus.get_serial_number(device)}
            for device in devices
        ]

    def _find_device(self, type_filter=None, serial=None):
        """Returns the first matching device or raises an error.

        :raise CanInitializationError:
            If not matching device could be found
        """
        if type_filter is not None:
            devices = ics.find_devices(type_filter)
        else:
            devices = ics.find_devices()

        for device in devices:
            if serial is None or self.get_serial_number(device) == str(serial):
                return device
        else:
            msg = ["No device"]

            if type_filter is not None:
                msg.append(f"with type {type_filter}")
            if serial is not None:
                msg.append(f"with serial {serial}")
            msg.append("found.")
            raise CanInitializationError(" ".join(msg))

    def _process_msg_queue(self, timeout=0.1):
        try:
            messages, errors = ics.get_messages(self.dev, False, timeout)
        except ics.RuntimeError:
            return
        for ics_msg in messages:
            if ics_msg.NetworkID not in self.channels:
                continue

            is_tx = bool(ics_msg.StatusBitField & ics.SPY_STATUS_TX_MSG)

            if is_tx:
                if bool(ics_msg.StatusBitField & ics.SPY_STATUS_GLOBAL_ERR):
                    continue

                receipt_key = (ics_msg.ArbIDOrHeader, ics_msg.DescriptionID)
                if ics_msg.DescriptionID and receipt_key in self.message_receipts:
                    self.message_receipts[receipt_key].set()
                if not self._receive_own_messages:
                    continue

            self.rx_buffer.append(ics_msg)
        if errors:
            logger.warning("%d error(s) found", errors)

            for msg, count in Counter(ics.get_error_messages(self.dev)).items():
                error = ICSApiError(*msg)
                if count > 1:
                    logger.warning(f"{error} (Repeated {count} times)")
                else:
                    logger.warning(error)

    def _get_timestamp_for_msg(self, ics_msg):
        if self._use_system_timestamp:
            # This is the system time stamp.
            # TimeSystem is loaded with the value received from the timeGetTime
            # call in the WIN32 multimedia API.
            #
            # The timeGetTime accuracy is up to 1 millisecond. See the WIN32
            # API documentation for more information.
            #
            # This timestamp is useful for time comparing with other system
            # events or data which is not synced with the neoVI timestamp.
            #
            # Currently, TimeSystem2 is not used.
            return ics_msg.TimeSystem
        else:
            # This is the hardware time stamp.
            return ics.get_timestamp_for_msg(self.dev, ics_msg)

    def _ics_msg_to_message(self, ics_msg):
        is_fd = ics_msg.Protocol == ics.SPY_PROTOCOL_CANFD

        if is_fd:
            if ics_msg.ExtraDataPtrEnabled:
                data = ics_msg.ExtraDataPtr[: ics_msg.NumberBytesData]
            else:
                data = ics_msg.Data[: ics_msg.NumberBytesData]

            return Message(
                timestamp=self._get_timestamp_for_msg(ics_msg),
                arbitration_id=ics_msg.ArbIDOrHeader,
                data=data,
                dlc=ics_msg.NumberBytesData,
                is_extended_id=bool(ics_msg.StatusBitField & ics.SPY_STATUS_XTD_FRAME),
                is_fd=is_fd,
                is_rx=not bool(ics_msg.StatusBitField & ics.SPY_STATUS_TX_MSG),
                is_remote_frame=bool(
                    ics_msg.StatusBitField & ics.SPY_STATUS_REMOTE_FRAME
                ),
                is_error_frame=bool(
                    ics_msg.StatusBitField2 & ics.SPY_STATUS2_ERROR_FRAME
                ),
                error_state_indicator=bool(
                    ics_msg.StatusBitField3 & ics.SPY_STATUS3_CANFD_ESI
                ),
                bitrate_switch=bool(
                    ics_msg.StatusBitField3 & ics.SPY_STATUS3_CANFD_BRS
                ),
                channel=ics_msg.NetworkID,
            )
        else:
            return Message(
                timestamp=self._get_timestamp_for_msg(ics_msg),
                arbitration_id=ics_msg.ArbIDOrHeader,
                data=ics_msg.Data[: ics_msg.NumberBytesData],
                dlc=ics_msg.NumberBytesData,
                is_extended_id=bool(ics_msg.StatusBitField & ics.SPY_STATUS_XTD_FRAME),
                is_fd=is_fd,
                is_rx=not bool(ics_msg.StatusBitField & ics.SPY_STATUS_TX_MSG),
                is_remote_frame=bool(
                    ics_msg.StatusBitField & ics.SPY_STATUS_REMOTE_FRAME
                ),
                is_error_frame=bool(
                    ics_msg.StatusBitField2 & ics.SPY_STATUS2_ERROR_FRAME
                ),
                channel=ics_msg.NetworkID,
            )

    def _recv_internal(self, timeout=0.1):
        if not self.rx_buffer:
            self._process_msg_queue(timeout=timeout)
        try:
            ics_msg = self.rx_buffer.popleft()
            msg = self._ics_msg_to_message(ics_msg)
        except IndexError:
            return None, False
        return msg, False

    def send(self, msg, timeout=0):
        """Transmit a message to the CAN bus.

        :param Message msg: A message object.

        :param float timeout:
            If > 0, wait up to this many seconds for message to be ACK'ed.
            If timeout is exceeded, an exception will be raised.
            None blocks indefinitely.

        :raises ValueError:
            if the message is invalid
        :raises can.CanTimeoutError:
            if sending timed out
        :raises CanOperationError:
            If the bus is closed or the message could otherwise not be sent.
            May or may not be a :class:`~ICSOperationError`.
        """
        if not ics.validate_hobject(self.dev):
            raise CanOperationError("bus not open")

        # Check for valid DLC to avoid passing extra long data to the driver
        if msg.is_fd:
            if msg.dlc > 64:
                raise ValueError(
                    f"DLC was {msg.dlc} but it should be <= 64 for CAN FD frames"
                )
        elif msg.dlc > 8:
            raise ValueError(
                f"DLC was {msg.dlc} but it should be <= 8 for normal CAN frames"
            )

        message = ics.SpyMessage()

        flag0 = 0
        if msg.is_extended_id:
            flag0 |= ics.SPY_STATUS_XTD_FRAME
        if msg.is_remote_frame:
            flag0 |= ics.SPY_STATUS_REMOTE_FRAME

        flag3 = 0
        if msg.is_fd:
            message.Protocol = ics.SPY_PROTOCOL_CANFD
            if msg.bitrate_switch:
                flag3 |= ics.SPY_STATUS3_CANFD_BRS
            if msg.error_state_indicator:
                flag3 |= ics.SPY_STATUS3_CANFD_ESI

        message.ArbIDOrHeader = msg.arbitration_id
        msg_data = msg.data[: msg.dlc]
        message.NumberBytesData = msg.dlc
        message.Data = tuple(msg_data[:8])
        if msg.is_fd and len(msg_data) > 8:
            message.ExtraDataPtrEnabled = 1
            message.ExtraDataPtr = tuple(msg_data)
        message.StatusBitField = flag0
        message.StatusBitField2 = 0
        message.StatusBitField3 = flag3
        if msg.channel is not None:
            message.NetworkID = msg.channel
        elif len(self.channels) == 1:
            message.NetworkID = self.channels[0]
        else:
            raise ValueError("msg.channel must be set when using multiple channels.")

        if timeout != 0:
            msg_desc_id = next(description_id)
            message.DescriptionID = msg_desc_id
            receipt_key = (msg.arbitration_id, msg_desc_id)
            self.message_receipts[receipt_key].clear()

        try:
            ics.transmit_messages(self.dev, message)
        except ics.RuntimeError:
            raise ICSOperationError(*ics.get_last_api_error(self.dev)) from None

        # If timeout is set, wait for ACK
        # This requires a notifier for the bus or
        # some other thread calling recv periodically
        if timeout != 0:
            got_receipt = self.message_receipts[receipt_key].wait(timeout)
            # We no longer need this receipt, so no point keeping it in memory
            del self.message_receipts[receipt_key]
            if not got_receipt:
                raise CanTimeoutError("Transmit timeout")



"""
Contains handling of MF4 logging files.

MF4 files represent Measurement Data Format (MDF) version 4 as specified by
the ASAM MDF standard (see https://www.asam.net/standards/detail/mdf/)
"""
import logging
from datetime import datetime
from hashlib import md5
from io import BufferedIOBase, BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Generator, Optional, Union, cast

from ..message import Message
from ..typechecking import StringPathLike
from ..util import channel2int, dlc2len, len2dlc
from .generic import FileIOMessageWriter, MessageReader

logger = logging.getLogger("can.io.mf4")

try:
    import asammdf
    import numpy as np
    from asammdf import Signal
    from asammdf.blocks.mdf_v4 import MDF4
    from asammdf.blocks.v4_blocks import SourceInformation
    from asammdf.blocks.v4_constants import BUS_TYPE_CAN, SOURCE_BUS
    from asammdf.mdf import MDF

    STD_DTYPE = np.dtype(
        [
            ("CAN_DataFrame.BusChannel", "<u1"),
            ("CAN_DataFrame.ID", "<u4"),
            ("CAN_DataFrame.IDE", "<u1"),
            ("CAN_DataFrame.DLC", "<u1"),
            ("CAN_DataFrame.DataLength", "<u1"),
            ("CAN_DataFrame.DataBytes", "(64,)u1"),
            ("CAN_DataFrame.Dir", "<u1"),
            ("CAN_DataFrame.EDL", "<u1"),
            ("CAN_DataFrame.BRS", "<u1"),
            ("CAN_DataFrame.ESI", "<u1"),
        ]
    )

    ERR_DTYPE = np.dtype(
        [
            ("CAN_ErrorFrame.BusChannel", "<u1"),
            ("CAN_ErrorFrame.ID", "<u4"),
            ("CAN_ErrorFrame.IDE", "<u1"),
            ("CAN_ErrorFrame.DLC", "<u1"),
            ("CAN_ErrorFrame.DataLength", "<u1"),
            ("CAN_ErrorFrame.DataBytes", "(64,)u1"),
            ("CAN_ErrorFrame.Dir", "<u1"),
            ("CAN_ErrorFrame.EDL", "<u1"),
            ("CAN_ErrorFrame.BRS", "<u1"),
            ("CAN_ErrorFrame.ESI", "<u1"),
        ]
    )

    RTR_DTYPE = np.dtype(
        [
            ("CAN_RemoteFrame.BusChannel", "<u1"),
            ("CAN_RemoteFrame.ID", "<u4"),
            ("CAN_RemoteFrame.IDE", "<u1"),
            ("CAN_RemoteFrame.DLC", "<u1"),
            ("CAN_RemoteFrame.DataLength", "<u1"),
            ("CAN_RemoteFrame.Dir", "<u1"),
        ]
    )
except ImportError:
    asammdf = None  # type: ignore


CAN_MSG_EXT = 0x80000000
CAN_ID_MASK = 0x1FFFFFFF


class MF4Writer(FileIOMessageWriter):
    """Logs CAN data to an ASAM Measurement Data File v4 (.mf4).

    MF4Writer does not support append mode.

    If a message has a timestamp smaller than the previous one or None,
    it gets assigned the timestamp that was written for the last message.
    It the first message does not have a timestamp, it is set to zero.
    """

    def __init__(
        self,
        file: Union[StringPathLike, BinaryIO],
        database: Optional[StringPathLike] = None,
        compression_level: int = 2,
        **kwargs: Any,
    ) -> None:
        """
        :param file:
            A path-like object or as file-like object to write to.
            If this is a file-like object, is has to be opened in
            binary write mode, not text write mode.
        :param database:
            optional path to a DBC or ARXML file that contains message description.
        :param compression_level:
            compression option as integer (default 2)
            * 0 - no compression
            * 1 - deflate (slower, but produces smaller files)
            * 2 - transposition + deflate (slowest, but produces the smallest files)
        """
        if asammdf is None:
            raise NotImplementedError(
                "The asammdf package was not found. Install python-can with "
                "the optional dependency [mf4] to use the MF4Writer."
            )

        if kwargs.get("append", False):
            raise ValueError(
                f"{self.__class__.__name__} is currently not equipped to "
                f"append messages to an existing file."
            )

        super().__init__(file, mode="w+b")
        now = datetime.now()
        self._mdf = cast(MDF4, MDF(version="4.10"))
        self._mdf.header.start_time = now
        self.last_timestamp = self._start_time = now.timestamp()

        self._compression_level = compression_level

        if database:
            database = Path(database).resolve()
            if database.exists():
                data = database.read_bytes()
                attachment = data, database.name, md5(data).digest()
            else:
                attachment = None
        else:
            attachment = None

        acquisition_source = SourceInformation(
            source_type=SOURCE_BUS, bus_type=BUS_TYPE_CAN
        )

        # standard frames group
        self._mdf.append(
            Signal(
                name="CAN_DataFrame",
                samples=np.array([], dtype=STD_DTYPE),
                timestamps=np.array([], dtype="<f8"),
                attachment=attachment,
                source=acquisition_source,
            )
        )

        # error frames group
        self._mdf.append(
            Signal(
                name="CAN_ErrorFrame",
                samples=np.array([], dtype=ERR_DTYPE),
                timestamps=np.array([], dtype="<f8"),
                attachment=attachment,
                source=acquisition_source,
            )
        )

        # remote frames group
        self._mdf.append(
            Signal(
                name="CAN_RemoteFrame",
                samples=np.array([], dtype=RTR_DTYPE),
                timestamps=np.array([], dtype="<f8"),
                attachment=attachment,
                source=acquisition_source,
            )
        )

        self._std_buffer = np.zeros(1, dtype=STD_DTYPE)
        self._err_buffer = np.zeros(1, dtype=ERR_DTYPE)
        self._rtr_buffer = np.zeros(1, dtype=RTR_DTYPE)

    def file_size(self) -> int:
        """Return an estimate of the current file size in bytes."""
        # TODO: find solution without accessing private attributes of asammdf
        return cast(int, self._mdf._tempfile.tell())  # pylint: disable=protected-access

    def stop(self) -> None:
        self._mdf.save(self.file, compression=self._compression_level)
        self._mdf.close()
        super().stop()

    def on_message_received(self, msg: Message) -> None:
        channel = channel2int(msg.channel)

        timestamp = msg.timestamp
        if timestamp is None:
            timestamp = self.last_timestamp
        else:
            self.last_timestamp = max(self.last_timestamp, timestamp)

        timestamp -= self._start_time

        if msg.is_remote_frame:
            if channel is not None:
                self._rtr_buffer["CAN_RemoteFrame.BusChannel"] = channel

            self._rtr_buffer["CAN_RemoteFrame.ID"] = msg.arbitration_id
            self._rtr_buffer["CAN_RemoteFrame.IDE"] = int(msg.is_extended_id)
            self._rtr_buffer["CAN_RemoteFrame.Dir"] = 0 if msg.is_rx else 1
            self._rtr_buffer["CAN_RemoteFrame.DLC"] = msg.dlc

            sigs = [(np.array([timestamp]), None), (self._rtr_buffer, None)]
            self._mdf.extend(2, sigs)

        elif msg.is_error_frame:
            if channel is not None:
                self._err_buffer["CAN_ErrorFrame.BusChannel"] = channel

            self._err_buffer["CAN_ErrorFrame.ID"] = msg.arbitration_id
            self._err_buffer["CAN_ErrorFrame.IDE"] = int(msg.is_extended_id)
            self._err_buffer["CAN_ErrorFrame.Dir"] = 0 if msg.is_rx else 1
            data = msg.data
            size = len(data)
            self._err_buffer["CAN_ErrorFrame.DataLength"] = size
            self._err_buffer["CAN_ErrorFrame.DataBytes"][0, :size] = data
            if msg.is_fd:
                self._err_buffer["CAN_ErrorFrame.DLC"] = len2dlc(msg.dlc)
                self._err_buffer["CAN_ErrorFrame.ESI"] = int(msg.error_state_indicator)
                self._err_buffer["CAN_ErrorFrame.BRS"] = int(msg.bitrate_switch)
                self._err_buffer["CAN_ErrorFrame.EDL"] = 1
            else:
                self._err_buffer["CAN_ErrorFrame.DLC"] = msg.dlc
                self._err_buffer["CAN_ErrorFrame.ESI"] = 0
                self._err_buffer["CAN_ErrorFrame.BRS"] = 0
                self._err_buffer["CAN_ErrorFrame.EDL"] = 0

            sigs = [(np.array([timestamp]), None), (self._err_buffer, None)]
            self._mdf.extend(1, sigs)

        else:
            if channel is not None:
                self._std_buffer["CAN_DataFrame.BusChannel"] = channel

            self._std_buffer["CAN_DataFrame.ID"] = msg.arbitration_id
            self._std_buffer["CAN_DataFrame.IDE"] = int(msg.is_extended_id)
            self._std_buffer["CAN_DataFrame.Dir"] = 0 if msg.is_rx else 1
            data = msg.data
            size = len(data)
            self._std_buffer["CAN_DataFrame.DataLength"] = size
            self._std_buffer["CAN_DataFrame.DataBytes"][0, :size] = data
            if msg.is_fd:
                self._std_buffer["CAN_DataFrame.DLC"] = len2dlc(msg.dlc)
                self._std_buffer["CAN_DataFrame.ESI"] = int(msg.error_state_indicator)
                self._std_buffer["CAN_DataFrame.BRS"] = int(msg.bitrate_switch)
                self._std_buffer["CAN_DataFrame.EDL"] = 1
            else:
                self._std_buffer["CAN_DataFrame.DLC"] = msg.dlc
                self._std_buffer["CAN_DataFrame.ESI"] = 0
                self._std_buffer["CAN_DataFrame.BRS"] = 0
                self._std_buffer["CAN_DataFrame.EDL"] = 0

            sigs = [(np.array([timestamp]), None), (self._std_buffer, None)]
            self._mdf.extend(0, sigs)

        # reset buffer structure
        self._std_buffer = np.zeros(1, dtype=STD_DTYPE)
        self._err_buffer = np.zeros(1, dtype=ERR_DTYPE)
        self._rtr_buffer = np.zeros(1, dtype=RTR_DTYPE)


class MF4Reader(MessageReader):
    """
    Iterator of CAN messages from a MF4 logging file.

    The MF4Reader only supports MF4 files that were recorded with python-can.
    """

    def __init__(self, file: Union[StringPathLike, BinaryIO]) -> None:
        """
        :param file: a path-like object or as file-like object to read from
                        If this is a file-like object, is has to be opened in
                        binary read mode, not text read mode.
        """
        if asammdf is None:
            raise NotImplementedError(
                "The asammdf package was not found. Install python-can with "
                "the optional dependency [mf4] to use the MF4Reader."
            )

        super().__init__(file, mode="rb")

        self._mdf: MDF4
        if isinstance(file, BufferedIOBase):
            self._mdf = MDF(BytesIO(file.read()))
        else:
            self._mdf = MDF(file)

        self.start_timestamp = self._mdf.header.start_time.timestamp()

        masters = [self._mdf.get_master(i) for i in range(3)]

        masters = [
            np.core.records.fromarrays((master, np.ones(len(master)) * i))
            for i, master in enumerate(masters)
        ]

        self.masters = np.sort(np.concatenate(masters))

    def __iter__(self) -> Generator[Message, None, None]:
        standard_counter = 0
        error_counter = 0
        rtr_counter = 0

        for timestamp, group_index in self.masters:
            # standard frames
            if group_index == 0:
                sample = self._mdf.get(
                    "CAN_DataFrame",
                    group=group_index,
                    raw=True,
                    record_offset=standard_counter,
                    record_count=1,
                )

                try:
                    channel = int(sample["CAN_DataFrame.BusChannel"][0])
                except ValueError:
                    channel = None

                if sample["CAN_DataFrame.EDL"] == 0:
                    is_extended_id = bool(sample["CAN_DataFrame.IDE"][0])
                    arbitration_id = int(sample["CAN_DataFrame.ID"][0])
                    is_rx = int(sample["CAN_DataFrame.Dir"][0]) == 0
                    size = int(sample["CAN_DataFrame.DataLength"][0])
                    dlc = int(sample["CAN_DataFrame.DLC"][0])
                    data = sample["CAN_DataFrame.DataBytes"][0, :size].tobytes()

                    msg = Message(
                        timestamp=timestamp + self.start_timestamp,
                        is_error_frame=False,
                        is_remote_frame=False,
                        is_fd=False,
                        is_extended_id=is_extended_id,
                        channel=channel,
                        is_rx=is_rx,
                        arbitration_id=arbitration_id,
                        data=data,
                        dlc=dlc,
                    )

                else:
                    is_extended_id = bool(sample["CAN_DataFrame.IDE"][0])
                    arbitration_id = int(sample["CAN_DataFrame.ID"][0])
                    is_rx = int(sample["CAN_DataFrame.Dir"][0]) == 0
                    size = int(sample["CAN_DataFrame.DataLength"][0])
                    dlc = dlc2len(sample["CAN_DataFrame.DLC"][0])
                    data = sample["CAN_DataFrame.DataBytes"][0, :size].tobytes()
                    error_state_indicator = bool(sample["CAN_DataFrame.ESI"][0])
                    bitrate_switch = bool(sample["CAN_DataFrame.BRS"][0])

                    msg = Message(
                        timestamp=timestamp + self.start_timestamp,
                        is_error_frame=False,
                        is_remote_frame=False,
                        is_fd=True,
                        is_extended_id=is_extended_id,
                        channel=channel,
                        arbitration_id=arbitration_id,
                        is_rx=is_rx,
                        data=data,
                        dlc=dlc,
                        bitrate_switch=bitrate_switch,
                        error_state_indicator=error_state_indicator,
                    )

                yield msg
                standard_counter += 1

            # error frames
            elif group_index == 1:
                sample = self._mdf.get(
                    "CAN_ErrorFrame",
                    group=group_index,
                    raw=True,
                    record_offset=error_counter,
                    record_count=1,
                )

                try:
                    channel = int(sample["CAN_ErrorFrame.BusChannel"][0])
                except ValueError:
                    channel = None

                if sample["CAN_ErrorFrame.EDL"] == 0:
                    is_extended_id = bool(sample["CAN_ErrorFrame.IDE"][0])
                    arbitration_id = int(sample["CAN_ErrorFrame.ID"][0])
                    is_rx = int(sample["CAN_ErrorFrame.Dir"][0]) == 0
                    size = int(sample["CAN_ErrorFrame.DataLength"][0])
                    dlc = int(sample["CAN_ErrorFrame.DLC"][0])
                    data = sample["CAN_ErrorFrame.DataBytes"][0, :size].tobytes()

                    msg = Message(
                        timestamp=timestamp + self.start_timestamp,
                        is_error_frame=True,
                        is_remote_frame=False,
                        is_fd=False,
                        is_extended_id=is_extended_id,
                        channel=channel,
                        arbitration_id=arbitration_id,
                        is_rx=is_rx,
                        data=data,
                        dlc=dlc,
                    )

                else:
                    is_extended_id = bool(sample["CAN_ErrorFrame.IDE"][0])
                    arbitration_id = int(sample["CAN_ErrorFrame.ID"][0])
                    is_rx = int(sample["CAN_ErrorFrame.Dir"][0]) == 0
                    size = int(sample["CAN_ErrorFrame.DataLength"][0])
                    dlc = dlc2len(sample["CAN_ErrorFrame.DLC"][0])
                    data = sample["CAN_ErrorFrame.DataBytes"][0, :size].tobytes()
                    error_state_indicator = bool(sample["CAN_ErrorFrame.ESI"][0])
                    bitrate_switch = bool(sample["CAN_ErrorFrame.BRS"][0])

                    msg = Message(
                        timestamp=timestamp + self.start_timestamp,
                        is_error_frame=True,
                        is_remote_frame=False,
                        is_fd=True,
                        is_extended_id=is_extended_id,
                        channel=channel,
                        arbitration_id=arbitration_id,
                        is_rx=is_rx,
                        data=data,
                        dlc=dlc,
                        bitrate_switch=bitrate_switch,
                        error_state_indicator=error_state_indicator,
                    )

                yield msg
                error_counter += 1

            # remote frames
            else:
                sample = self._mdf.get(
                    "CAN_RemoteFrame",
                    group=group_index,
                    raw=True,
                    record_offset=rtr_counter,
                    record_count=1,
                )

                try:
                    channel = int(sample["CAN_RemoteFrame.BusChannel"][0])
                except ValueError:
                    channel = None

                is_extended_id = bool(sample["CAN_RemoteFrame.IDE"][0])
                arbitration_id = int(sample["CAN_RemoteFrame.ID"][0])
                is_rx = int(sample["CAN_RemoteFrame.Dir"][0]) == 0
                dlc = int(sample["CAN_RemoteFrame.DLC"][0])

                msg = Message(
                    timestamp=timestamp + self.start_timestamp,
                    is_error_frame=False,
                    is_remote_frame=True,
                    is_fd=False,
                    is_extended_id=is_extended_id,
                    channel=channel,
                    arbitration_id=arbitration_id,
                    is_rx=is_rx,
                    dlc=dlc,
                )

                yield msg

                rtr_counter += 1

        self.stop()

    def stop(self) -> None:
        self._mdf.close()
        super().stop()



# pylint: disable=too-many-lines
import math
from typing import Iterator, List, Mapping, cast

from can.typechecking import BitTimingDict, BitTimingFdDict


class BitTiming(Mapping):
    """Representation of a bit timing configuration for a CAN 2.0 bus.

    The class can be constructed in multiple ways, depending on the information
    available. The preferred way is using CAN clock frequency, prescaler, tseg1, tseg2 and sjw::

        can.BitTiming(f_clock=8_000_000, brp=1, tseg1=5, tseg2=1, sjw=1)

    Alternatively you can set the bitrate instead of the bit rate prescaler::

        can.BitTiming.from_bitrate_and_segments(
            f_clock=8_000_000, bitrate=1_000_000, tseg1=5, tseg2=1, sjw=1
        )

    It is also possible to specify BTR registers::

        can.BitTiming.from_registers(f_clock=8_000_000, btr0=0x00, btr1=0x14)

    or to calculate the timings for a given sample point::

        can.BitTiming.from_sample_point(f_clock=8_000_000, bitrate=1_000_000, sample_point=75.0)
    """

    def __init__(
        self,
        f_clock: int,
        brp: int,
        tseg1: int,
        tseg2: int,
        sjw: int,
        nof_samples: int = 1,
    ) -> None:
        """
        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int brp:
            Bit rate prescaler.
        :param int tseg1:
            Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
        :param int tseg2:
            Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int sjw:
            The Synchronization Jump Width. Decides the maximum number of time quanta
            that the controller can resynchronize every bit.
        :param int nof_samples:
            Either 1 or 3. Some CAN controllers can also sample each bit three times.
            In this case, the bit will be sampled three quanta in a row,
            with the last sample being taken in the edge between TSEG1 and TSEG2.
            Three samples should only be used for relatively slow baudrates.
        :raises ValueError:
            if the arguments are invalid.
        """
        self._data: BitTimingDict = {
            "f_clock": f_clock,
            "brp": brp,
            "tseg1": tseg1,
            "tseg2": tseg2,
            "sjw": sjw,
            "nof_samples": nof_samples,
        }
        self._validate()

    def _validate(self) -> None:
        if not 8 <= self.nbt <= 25:
            raise ValueError(f"nominal bit time (={self.nbt}) must be in [8...25].")

        if not 1 <= self.brp <= 64:
            raise ValueError(f"bitrate prescaler (={self.brp}) must be in [1...64].")

        if not 5_000 <= self.bitrate <= 2_000_000:
            raise ValueError(
                f"bitrate (={self.bitrate}) must be in [5,000...2,000,000]."
            )

        if not 1 <= self.tseg1 <= 16:
            raise ValueError(f"tseg1 (={self.tseg1}) must be in [1...16].")

        if not 1 <= self.tseg2 <= 8:
            raise ValueError(f"tseg2 (={self.tseg2}) must be in [1...8].")

        if not 1 <= self.sjw <= 4:
            raise ValueError(f"sjw (={self.sjw}) must be in [1...4].")

        if self.sjw > self.tseg2:
            raise ValueError(
                f"sjw (={self.sjw}) must not be greater than tseg2 (={self.tseg2})."
            )

        if self.sample_point < 50.0:
            raise ValueError(
                f"The sample point must be greater than or equal to 50% "
                f"(sample_point={self.sample_point:.2f}%)."
            )

        if self.nof_samples not in (1, 3):
            raise ValueError("nof_samples must be 1 or 3")

    @classmethod
    def from_bitrate_and_segments(
        cls,
        f_clock: int,
        bitrate: int,
        tseg1: int,
        tseg2: int,
        sjw: int,
        nof_samples: int = 1,
    ) -> "BitTiming":
        """Create a :class:`~can.BitTiming` instance from bitrate and segment lengths.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int bitrate:
            Bitrate in bit/s.
        :param int tseg1:
            Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
        :param int tseg2:
            Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int sjw:
            The Synchronization Jump Width. Decides the maximum number of time quanta
            that the controller can resynchronize every bit.
        :param int nof_samples:
            Either 1 or 3. Some CAN controllers can also sample each bit three times.
            In this case, the bit will be sampled three quanta in a row,
            with the last sample being taken in the edge between TSEG1 and TSEG2.
            Three samples should only be used for relatively slow baudrates.
        :raises ValueError:
            if the arguments are invalid.
        """
        try:
            brp = int(round(f_clock / (bitrate * (1 + tseg1 + tseg2))))
        except ZeroDivisionError:
            raise ValueError("Invalid inputs") from None

        bt = cls(
            f_clock=f_clock,
            brp=brp,
            tseg1=tseg1,
            tseg2=tseg2,
            sjw=sjw,
            nof_samples=nof_samples,
        )
        if abs(bt.bitrate - bitrate) > bitrate / 256:
            raise ValueError(
                f"the effective bitrate (={bt.bitrate}) diverges "
                f"from the requested bitrate (={bitrate})"
            )
        return bt

    @classmethod
    def from_registers(
        cls,
        f_clock: int,
        btr0: int,
        btr1: int,
    ) -> "BitTiming":
        """Create a :class:`~can.BitTiming` instance from registers btr0 and btr1.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int btr0:
            The BTR0 register value used by many CAN controllers.
        :param int btr1:
            The BTR1 register value used by many CAN controllers.
        :raises ValueError:
            if the arguments are invalid.
        """
        brp = (btr0 & 0x3F) + 1
        sjw = (btr0 >> 6) + 1
        tseg1 = (btr1 & 0xF) + 1
        tseg2 = ((btr1 >> 4) & 0x7) + 1
        nof_samples = 3 if btr1 & 0x80 else 1
        return cls(
            brp=brp,
            f_clock=f_clock,
            tseg1=tseg1,
            tseg2=tseg2,
            sjw=sjw,
            nof_samples=nof_samples,
        )

    @classmethod
    def from_sample_point(
        cls, f_clock: int, bitrate: int, sample_point: float = 69.0
    ) -> "BitTiming":
        """Create a :class:`~can.BitTiming` instance for a sample point.

        This function tries to find bit timings, which are close to the requested
        sample point. It does not take physical bus properties into account, so the
        calculated bus timings might not work properly for you.

        The :func:`oscillator_tolerance` function might be helpful to evaluate the
        bus timings.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int bitrate:
            Bitrate in bit/s.
        :param int sample_point:
            The sample point value in percent.
        :raises ValueError:
            if the arguments are invalid.
        """

        if sample_point < 50.0:
            raise ValueError(f"sample_point (={sample_point}) must not be below 50%.")

        possible_solutions: List[BitTiming] = []
        for brp in range(1, 65):
            nbt = round(int(f_clock / (bitrate * brp)))
            if nbt < 8:
                break

            effective_bitrate = f_clock / (nbt * brp)
            if abs(effective_bitrate - bitrate) > bitrate / 256:
                continue

            tseg1 = int(round(sample_point / 100 * nbt)) - 1
            # limit tseg1, so tseg2 is at least 1 TQ
            tseg1 = min(tseg1, nbt - 2)

            tseg2 = nbt - tseg1 - 1
            sjw = min(tseg2, 4)

            try:
                bt = BitTiming(
                    f_clock=f_clock,
                    brp=brp,
                    tseg1=tseg1,
                    tseg2=tseg2,
                    sjw=sjw,
                )
                possible_solutions.append(bt)
            except ValueError:
                continue

        if not possible_solutions:
            raise ValueError("No suitable bit timings found.")

        # sort solutions
        for key, reverse in (
            # prefer low prescaler
            (lambda x: x.brp, False),
            # prefer low sample point deviation from requested values
            (lambda x: abs(x.sample_point - sample_point), False),
        ):
            possible_solutions.sort(key=key, reverse=reverse)

        return possible_solutions[0]

    @property
    def f_clock(self) -> int:
        """The CAN system clock frequency in Hz."""
        return self._data["f_clock"]

    @property
    def bitrate(self) -> int:
        """Bitrate in bits/s."""
        return int(round(self.f_clock / (self.nbt * self.brp)))

    @property
    def brp(self) -> int:
        """Bit Rate Prescaler."""
        return self._data["brp"]

    @property
    def tq(self) -> int:
        """Time quantum in nanoseconds"""
        return int(round(self.brp / self.f_clock * 1e9))

    @property
    def nbt(self) -> int:
        """Nominal Bit Time."""
        return 1 + self.tseg1 + self.tseg2

    @property
    def tseg1(self) -> int:
        """Time segment 1.

        The number of quanta from (but not including) the Sync Segment to the sampling point.
        """
        return self._data["tseg1"]

    @property
    def tseg2(self) -> int:
        """Time segment 2.

        The number of quanta from the sampling point to the end of the bit.
        """
        return self._data["tseg2"]

    @property
    def sjw(self) -> int:
        """Synchronization Jump Width."""
        return self._data["sjw"]

    @property
    def nof_samples(self) -> int:
        """Number of samples (1 or 3)."""
        return self._data["nof_samples"]

    @property
    def sample_point(self) -> float:
        """Sample point in percent."""
        return 100.0 * (1 + self.tseg1) / (1 + self.tseg1 + self.tseg2)

    @property
    def btr0(self) -> int:
        """Bit timing register 0."""
        return (self.sjw - 1) << 6 | self.brp - 1

    @property
    def btr1(self) -> int:
        """Bit timing register 1."""
        sam = 1 if self.nof_samples == 3 else 0
        return sam << 7 | (self.tseg2 - 1) << 4 | self.tseg1 - 1

    def oscillator_tolerance(
        self,
        node_loop_delay_ns: float = 250.0,
        bus_length_m: float = 10.0,
    ) -> float:
        """Oscillator tolerance in percent according to ISO 11898-1.

        :param float node_loop_delay_ns:
            Transceiver loop delay in nanoseconds.
        :param float bus_length_m:
            Bus length in meters.
        """
        delay_per_meter = 5
        bidirectional_propagation_delay_ns = 2 * (
            node_loop_delay_ns + delay_per_meter * bus_length_m
        )

        prop_seg = math.ceil(bidirectional_propagation_delay_ns / self.tq)
        nom_phase_seg1 = self.tseg1 - prop_seg
        nom_phase_seg2 = self.tseg2
        df_clock_list = [
            _oscillator_tolerance_condition_1(nom_sjw=self.sjw, nbt=self.nbt),
            _oscillator_tolerance_condition_2(
                nbt=self.nbt,
                nom_phase_seg1=nom_phase_seg1,
                nom_phase_seg2=nom_phase_seg2,
            ),
        ]
        return max(0.0, min(df_clock_list) * 100)

    def recreate_with_f_clock(self, f_clock: int) -> "BitTiming":
        """Return a new :class:`~can.BitTiming` instance with the given *f_clock* but the same
        bit rate and sample point.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :raises ValueError:
            if no suitable bit timings were found.
        """
        # try the most simple solution first: another bitrate prescaler
        try:
            return BitTiming.from_bitrate_and_segments(
                f_clock=f_clock,
                bitrate=self.bitrate,
                tseg1=self.tseg1,
                tseg2=self.tseg2,
                sjw=self.sjw,
                nof_samples=self.nof_samples,
            )
        except ValueError:
            pass

        # create a new timing instance with the same sample point
        bt = BitTiming.from_sample_point(
            f_clock=f_clock, bitrate=self.bitrate, sample_point=self.sample_point
        )
        if abs(bt.sample_point - self.sample_point) > 1.0:
            raise ValueError(
                "f_clock change failed because of sample point discrepancy."
            )
        # adapt synchronization jump width, so it has the same size relative to bit time as self
        sjw = int(round(self.sjw / self.nbt * bt.nbt))
        sjw = max(1, min(4, bt.tseg2, sjw))
        bt._data["sjw"] = sjw  # pylint: disable=protected-access
        bt._data["nof_samples"] = self.nof_samples  # pylint: disable=protected-access
        bt._validate()  # pylint: disable=protected-access
        return bt

    def __str__(self) -> str:
        segments = [
            f"BR: {self.bitrate:_} bit/s",
            f"SP: {self.sample_point:.2f}%",
            f"BRP: {self.brp}",
            f"TSEG1: {self.tseg1}",
            f"TSEG2: {self.tseg2}",
            f"SJW: {self.sjw}",
            f"BTR: {self.btr0:02X}{self.btr1:02X}h",
            f"CLK: {self.f_clock / 1e6:.0f}MHz",
        ]
        return ", ".join(segments)

    def __repr__(self) -> str:
        args = ", ".join(f"{key}={value}" for key, value in self.items())
        return f"can.{self.__class__.__name__}({args})"

    def __getitem__(self, key: str) -> int:
        return cast(int, self._data.__getitem__(key))

    def __len__(self) -> int:
        return self._data.__len__()

    def __iter__(self) -> Iterator[str]:
        return self._data.__iter__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BitTiming):
            return False

        return self._data == other._data

    def __hash__(self) -> int:
        return tuple(self._data.values()).__hash__()


class BitTimingFd(Mapping):
    """Representation of a bit timing configuration for a CAN FD bus.

    The class can be constructed in multiple ways, depending on the information
    available. The preferred way is using CAN clock frequency, bit rate prescaler, tseg1,
    tseg2 and sjw for both the arbitration (nominal) and data phase::

        can.BitTimingFd(
            f_clock=80_000_000,
            nom_brp=1,
            nom_tseg1=59,
            nom_tseg2=20,
            nom_sjw=10,
            data_brp=1,
            data_tseg1=6,
            data_tseg2=3,
            data_sjw=2,
        )

    Alternatively you can set the bit rates instead of the bit rate prescalers::

        can.BitTimingFd.from_bitrate_and_segments(
            f_clock=80_000_000,
            nom_bitrate=1_000_000,
            nom_tseg1=59,
            nom_tseg2=20,
            nom_sjw=10,
            data_bitrate=8_000_000,
            data_tseg1=6,
            data_tseg2=3,
            data_sjw=2,
        )

    It is also possible to calculate the timings for a given
    pair of arbitration and data sample points::

        can.BitTimingFd.from_sample_point(
            f_clock=80_000_000,
            nom_bitrate=1_000_000,
            nom_sample_point=75.0,
            data_bitrate=8_000_000,
            data_sample_point=70.0,
        )
    """

    def __init__(
        self,
        f_clock: int,
        nom_brp: int,
        nom_tseg1: int,
        nom_tseg2: int,
        nom_sjw: int,
        data_brp: int,
        data_tseg1: int,
        data_tseg2: int,
        data_sjw: int,
    ) -> None:
        """
        Initialize a BitTimingFd instance with the specified parameters.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int nom_brp:
            Nominal (arbitration) phase bitrate prescaler.
        :param int nom_tseg1:
            Nominal phase Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
        :param int nom_tseg2:
            Nominal phase Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int nom_sjw:
            The Synchronization Jump Width for the nominal phase. This value determines
            the maximum number of time quanta that the controller can resynchronize every bit.
        :param int data_brp:
            Data phase bitrate prescaler.
        :param int data_tseg1:
            Data phase Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
        :param int data_tseg2:
            Data phase Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int data_sjw:
            The Synchronization Jump Width for the data phase. This value determines
            the maximum number of time quanta that the controller can resynchronize every bit.
        :raises ValueError:
            if the arguments are invalid.
        """
        self._data: BitTimingFdDict = {
            "f_clock": f_clock,
            "nom_brp": nom_brp,
            "nom_tseg1": nom_tseg1,
            "nom_tseg2": nom_tseg2,
            "nom_sjw": nom_sjw,
            "data_brp": data_brp,
            "data_tseg1": data_tseg1,
            "data_tseg2": data_tseg2,
            "data_sjw": data_sjw,
        }
        self._validate()

    def _validate(self) -> None:
        if self.nbt < 8:
            raise ValueError(f"nominal bit time (={self.nbt}) must be at least 8.")

        if self.dbt < 8:
            raise ValueError(f"data bit time (={self.dbt}) must be at least 8.")

        if not 1 <= self.nom_brp <= 256:
            raise ValueError(
                f"nominal bitrate prescaler (={self.nom_brp}) must be in [1...256]."
            )

        if not 1 <= self.data_brp <= 256:
            raise ValueError(
                f"data bitrate prescaler (={self.data_brp}) must be in [1...256]."
            )

        if not 5_000 <= self.nom_bitrate <= 2_000_000:
            raise ValueError(
                f"nom_bitrate (={self.nom_bitrate}) must be in [5,000...2,000,000]."
            )

        if not 25_000 <= self.data_bitrate <= 8_000_000:
            raise ValueError(
                f"data_bitrate (={self.data_bitrate}) must be in [25,000...8,000,000]."
            )

        if self.data_bitrate < self.nom_bitrate:
            raise ValueError(
                f"data_bitrate (={self.data_bitrate}) must be greater than or "
                f"equal to nom_bitrate (={self.nom_bitrate})"
            )

        if not 2 <= self.nom_tseg1 <= 256:
            raise ValueError(f"nom_tseg1 (={self.nom_tseg1}) must be in [2...256].")

        if not 1 <= self.nom_tseg2 <= 128:
            raise ValueError(f"nom_tseg2 (={self.nom_tseg2}) must be in [1...128].")

        if not 1 <= self.data_tseg1 <= 32:
            raise ValueError(f"data_tseg1 (={self.data_tseg1}) must be in [1...32].")

        if not 1 <= self.data_tseg2 <= 16:
            raise ValueError(f"data_tseg2 (={self.data_tseg2}) must be in [1...16].")

        if not 1 <= self.nom_sjw <= 128:
            raise ValueError(f"nom_sjw (={self.nom_sjw}) must be in [1...128].")

        if self.nom_sjw > self.nom_tseg2:
            raise ValueError(
                f"nom_sjw (={self.nom_sjw}) must not be "
                f"greater than nom_tseg2 (={self.nom_tseg2})."
            )

        if not 1 <= self.data_sjw <= 16:
            raise ValueError(f"data_sjw (={self.data_sjw}) must be in [1...128].")

        if self.data_sjw > self.data_tseg2:
            raise ValueError(
                f"data_sjw (={self.data_sjw}) must not be "
                f"greater than data_tseg2 (={self.data_tseg2})."
            )

        if self.nom_sample_point < 50.0:
            raise ValueError(
                f"The arbitration sample point must be greater than or equal to 50% "
                f"(nom_sample_point={self.nom_sample_point:.2f}%)."
            )

        if self.data_sample_point < 50.0:
            raise ValueError(
                f"The data sample point must be greater than or equal to 50% "
                f"(data_sample_point={self.data_sample_point:.2f}%)."
            )

    @classmethod
    def from_bitrate_and_segments(
        cls,
        f_clock: int,
        nom_bitrate: int,
        nom_tseg1: int,
        nom_tseg2: int,
        nom_sjw: int,
        data_bitrate: int,
        data_tseg1: int,
        data_tseg2: int,
        data_sjw: int,
    ) -> "BitTimingFd":
        """
        Create a :class:`~can.BitTimingFd` instance with the bitrates and segments lengths.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int nom_bitrate:
            Nominal (arbitration) phase bitrate in bit/s.
        :param int nom_tseg1:
            Nominal phase Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
        :param int nom_tseg2:
            Nominal phase Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int nom_sjw:
            The Synchronization Jump Width for the nominal phase. This value determines
            the maximum number of time quanta that the controller can resynchronize every bit.
        :param int data_bitrate:
            Data phase bitrate in bit/s.
        :param int data_tseg1:
            Data phase Time segment 1, that is, the number of quanta from (but not including)
            the Sync Segment to the sampling point.
        :param int data_tseg2:
            Data phase Time segment 2, that is, the number of quanta from the sampling
            point to the end of the bit.
        :param int data_sjw:
            The Synchronization Jump Width for the data phase. This value determines
            the maximum number of time quanta that the controller can resynchronize every bit.
        :raises ValueError:
            if the arguments are invalid.
        """
        try:
            nom_brp = int(round(f_clock / (nom_bitrate * (1 + nom_tseg1 + nom_tseg2))))
            data_brp = int(
                round(f_clock / (data_bitrate * (1 + data_tseg1 + data_tseg2)))
            )
        except ZeroDivisionError:
            raise ValueError("Invalid inputs.") from None

        bt = cls(
            f_clock=f_clock,
            nom_brp=nom_brp,
            nom_tseg1=nom_tseg1,
            nom_tseg2=nom_tseg2,
            nom_sjw=nom_sjw,
            data_brp=data_brp,
            data_tseg1=data_tseg1,
            data_tseg2=data_tseg2,
            data_sjw=data_sjw,
        )

        if abs(bt.nom_bitrate - nom_bitrate) > nom_bitrate / 256:
            raise ValueError(
                f"the effective nom. bitrate (={bt.nom_bitrate}) diverges "
                f"from the requested nom. bitrate (={nom_bitrate})"
            )

        if abs(bt.data_bitrate - data_bitrate) > data_bitrate / 256:
            raise ValueError(
                f"the effective data bitrate (={bt.data_bitrate}) diverges "
                f"from the requested data bitrate (={data_bitrate})"
            )

        return bt

    @classmethod
    def from_sample_point(
        cls,
        f_clock: int,
        nom_bitrate: int,
        nom_sample_point: float,
        data_bitrate: int,
        data_sample_point: float,
    ) -> "BitTimingFd":
        """Create a :class:`~can.BitTimingFd` instance for a given nominal/data sample point pair.

        This function tries to find bit timings, which are close to the requested
        sample points. It does not take physical bus properties into account, so the
        calculated bus timings might not work properly for you.

        The :func:`oscillator_tolerance` function might be helpful to evaluate the
        bus timings.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :param int nom_bitrate:
            Nominal bitrate in bit/s.
        :param int nom_sample_point:
            The sample point value of the arbitration phase in percent.
        :param int data_bitrate:
            Data bitrate in bit/s.
        :param int data_sample_point:
            The sample point value of the data phase in percent.
        :raises ValueError:
            if the arguments are invalid.
        """
        if nom_sample_point < 50.0:
            raise ValueError(
                f"nom_sample_point (={nom_sample_point}) must not be below 50%."
            )

        if data_sample_point < 50.0:
            raise ValueError(
                f"data_sample_point (={data_sample_point}) must not be below 50%."
            )

        possible_solutions: List[BitTimingFd] = []

        for nom_brp in range(1, 257):
            nbt = round(int(f_clock / (nom_bitrate * nom_brp)))
            if nbt < 8:
                break

            effective_nom_bitrate = f_clock / (nbt * nom_brp)
            if abs(effective_nom_bitrate - nom_bitrate) > nom_bitrate / 256:
                continue

            nom_tseg1 = int(round(nom_sample_point / 100 * nbt)) - 1
            # limit tseg1, so tseg2 is at least 1 TQ
            nom_tseg1 = min(nom_tseg1, nbt - 2)
            nom_tseg2 = nbt - nom_tseg1 - 1

            nom_sjw = min(nom_tseg2, 128)

            for data_brp in range(1, 257):
                dbt = round(int(f_clock / (data_bitrate * data_brp)))
                if dbt < 8:
                    break

                effective_data_bitrate = f_clock / (dbt * data_brp)
                if abs(effective_data_bitrate - data_bitrate) > data_bitrate / 256:
                    continue

                data_tseg1 = int(round(data_sample_point / 100 * dbt)) - 1
                # limit tseg1, so tseg2 is at least 1 TQ
                data_tseg1 = min(data_tseg1, dbt - 2)
                data_tseg2 = dbt - data_tseg1 - 1

                data_sjw = min(data_tseg2, 16)

                try:
                    bt = BitTimingFd(
                        f_clock=f_clock,
                        nom_brp=nom_brp,
                        nom_tseg1=nom_tseg1,
                        nom_tseg2=nom_tseg2,
                        nom_sjw=nom_sjw,
                        data_brp=data_brp,
                        data_tseg1=data_tseg1,
                        data_tseg2=data_tseg2,
                        data_sjw=data_sjw,
                    )
                    possible_solutions.append(bt)
                except ValueError:
                    continue

        if not possible_solutions:
            raise ValueError("No suitable bit timings found.")

        # prefer using the same prescaler for arbitration and data phase
        same_prescaler = list(
            filter(lambda x: x.nom_brp == x.data_brp, possible_solutions)
        )
        if same_prescaler:
            possible_solutions = same_prescaler

        # sort solutions
        for key, reverse in (
            # prefer low prescaler
            (lambda x: x.nom_brp + x.data_brp, False),
            # prefer same prescaler for arbitration and data
            (lambda x: abs(x.nom_brp - x.data_brp), False),
            # prefer low sample point deviation from requested values
            (
                lambda x: (
                    abs(x.nom_sample_point - nom_sample_point)
                    + abs(x.data_sample_point - data_sample_point)
                ),
                False,
            ),
        ):
            possible_solutions.sort(key=key, reverse=reverse)

        return possible_solutions[0]

    @property
    def f_clock(self) -> int:
        """The CAN system clock frequency in Hz."""
        return self._data["f_clock"]

    @property
    def nom_bitrate(self) -> int:
        """Nominal (arbitration phase) bitrate."""
        return int(round(self.f_clock / (self.nbt * self.nom_brp)))

    @property
    def nom_brp(self) -> int:
        """Prescaler value for the arbitration phase."""
        return self._data["nom_brp"]

    @property
    def nom_tq(self) -> int:
        """Nominal time quantum in nanoseconds"""
        return int(round(self.nom_brp / self.f_clock * 1e9))

    @property
    def nbt(self) -> int:
        """Number of time quanta in a bit of the arbitration phase."""
        return 1 + self.nom_tseg1 + self.nom_tseg2

    @property
    def nom_tseg1(self) -> int:
        """Time segment 1 value of the arbitration phase.

        This is the sum of the propagation time segment and the phase buffer segment 1.
        """
        return self._data["nom_tseg1"]

    @property
    def nom_tseg2(self) -> int:
        """Time segment 2 value of the arbitration phase. Also known as phase buffer segment 2."""
        return self._data["nom_tseg2"]

    @property
    def nom_sjw(self) -> int:
        """Synchronization jump width of the arbitration phase.

        The phase buffer segments may be shortened or lengthened by this value.
        """
        return self._data["nom_sjw"]

    @property
    def nom_sample_point(self) -> float:
        """Sample point of the arbitration phase in percent."""
        return 100.0 * (1 + self.nom_tseg1) / (1 + self.nom_tseg1 + self.nom_tseg2)

    @property
    def data_bitrate(self) -> int:
        """Bitrate of the data phase in bit/s."""
        return int(round(self.f_clock / (self.dbt * self.data_brp)))

    @property
    def data_brp(self) -> int:
        """Prescaler value for the data phase."""
        return self._data["data_brp"]

    @property
    def data_tq(self) -> int:
        """Data time quantum in nanoseconds"""
        return int(round(self.data_brp / self.f_clock * 1e9))

    @property
    def dbt(self) -> int:
        """Number of time quanta in a bit of the data phase."""
        return 1 + self.data_tseg1 + self.data_tseg2

    @property
    def data_tseg1(self) -> int:
        """TSEG1 value of the data phase.

        This is the sum of the propagation time segment and the phase buffer segment 1.
        """
        return self._data["data_tseg1"]

    @property
    def data_tseg2(self) -> int:
        """TSEG2 value of the data phase. Also known as phase buffer segment 2."""
        return self._data["data_tseg2"]

    @property
    def data_sjw(self) -> int:
        """Synchronization jump width of the data phase.

        The phase buffer segments may be shortened or lengthened by this value.
        """
        return self._data["data_sjw"]

    @property
    def data_sample_point(self) -> float:
        """Sample point of the data phase in percent."""
        return 100.0 * (1 + self.data_tseg1) / (1 + self.data_tseg1 + self.data_tseg2)

    def oscillator_tolerance(
        self,
        node_loop_delay_ns: float = 250.0,
        bus_length_m: float = 10.0,
    ) -> float:
        """Oscillator tolerance in percent according to ISO 11898-1.

        :param float node_loop_delay_ns:
            Transceiver loop delay in nanoseconds.
        :param float bus_length_m:
            Bus length in meters.
        """
        delay_per_meter = 5
        bidirectional_propagation_delay_ns = 2 * (
            node_loop_delay_ns + delay_per_meter * bus_length_m
        )

        prop_seg = math.ceil(bidirectional_propagation_delay_ns / self.nom_tq)
        nom_phase_seg1 = self.nom_tseg1 - prop_seg
        nom_phase_seg2 = self.nom_tseg2

        data_phase_seg2 = self.data_tseg2

        df_clock_list = [
            _oscillator_tolerance_condition_1(nom_sjw=self.nom_sjw, nbt=self.nbt),
            _oscillator_tolerance_condition_2(
                nbt=self.nbt,
                nom_phase_seg1=nom_phase_seg1,
                nom_phase_seg2=nom_phase_seg2,
            ),
            _oscillator_tolerance_condition_3(data_sjw=self.data_sjw, dbt=self.dbt),
            _oscillator_tolerance_condition_4(
                nom_phase_seg1=nom_phase_seg1,
                nom_phase_seg2=nom_phase_seg2,
                data_phase_seg2=data_phase_seg2,
                nbt=self.nbt,
                dbt=self.dbt,
                data_brp=self.data_brp,
                nom_brp=self.nom_brp,
            ),
            _oscillator_tolerance_condition_5(
                data_sjw=self.data_sjw,
                data_brp=self.data_brp,
                nom_brp=self.nom_brp,
                data_phase_seg2=data_phase_seg2,
                nom_phase_seg2=nom_phase_seg2,
                nbt=self.nbt,
                dbt=self.dbt,
            ),
        ]
        return max(0.0, min(df_clock_list) * 100)

    def recreate_with_f_clock(self, f_clock: int) -> "BitTimingFd":
        """Return a new :class:`~can.BitTimingFd` instance with the given *f_clock* but the same
        bit rates and sample points.

        :param int f_clock:
            The CAN system clock frequency in Hz.
        :raises ValueError:
            if no suitable bit timings were found.
        """
        # try the most simple solution first: another bitrate prescaler
        try:
            return BitTimingFd.from_bitrate_and_segments(
                f_clock=f_clock,
                nom_bitrate=self.nom_bitrate,
                nom_tseg1=self.nom_tseg1,
                nom_tseg2=self.nom_tseg2,
                nom_sjw=self.nom_sjw,
                data_bitrate=self.data_bitrate,
                data_tseg1=self.data_tseg1,
                data_tseg2=self.data_tseg2,
                data_sjw=self.data_sjw,
            )
        except ValueError:
            pass

        # create a new timing instance with the same sample points
        bt = BitTimingFd.from_sample_point(
            f_clock=f_clock,
            nom_bitrate=self.nom_bitrate,
            nom_sample_point=self.nom_sample_point,
            data_bitrate=self.data_bitrate,
            data_sample_point=self.data_sample_point,
        )
        if (
            abs(bt.nom_sample_point - self.nom_sample_point) > 1.0
            or abs(bt.data_sample_point - self.data_sample_point) > 1.0
        ):
            raise ValueError(
                "f_clock change failed because of sample point discrepancy."
            )
        # adapt synchronization jump width, so it has the same size relative to bit time as self
        nom_sjw = int(round(self.nom_sjw / self.nbt * bt.nbt))
        nom_sjw = max(1, min(bt.nom_tseg2, nom_sjw))
        bt._data["nom_sjw"] = nom_sjw  # pylint: disable=protected-access
        data_sjw = int(round(self.data_sjw / self.dbt * bt.dbt))
        data_sjw = max(1, min(bt.data_tseg2, data_sjw))
        bt._data["data_sjw"] = data_sjw  # pylint: disable=protected-access
        bt._validate()  # pylint: disable=protected-access
        return bt

    def __str__(self) -> str:
        segments = [
            f"NBR: {self.nom_bitrate:_} bit/s",
            f"NSP: {self.nom_sample_point:.2f}%",
            f"NBRP: {self.nom_brp}",
            f"NTSEG1: {self.nom_tseg1}",
            f"NTSEG2: {self.nom_tseg2}",
            f"NSJW: {self.nom_sjw}",
            f"DBR: {self.data_bitrate:_} bit/s",
            f"DSP: {self.data_sample_point:.2f}%",
            f"DBRP: {self.data_brp}",
            f"DTSEG1: {self.data_tseg1}",
            f"DTSEG2: {self.data_tseg2}",
            f"DSJW: {self.data_sjw}",
            f"CLK: {self.f_clock / 1e6:.0f}MHz",
        ]
        return ", ".join(segments)

    def __repr__(self) -> str:
        args = ", ".join(f"{key}={value}" for key, value in self.items())
        return f"can.{self.__class__.__name__}({args})"

    def __getitem__(self, key: str) -> int:
        return cast(int, self._data.__getitem__(key))

    def __len__(self) -> int:
        return self._data.__len__()

    def __iter__(self) -> Iterator[str]:
        return self._data.__iter__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BitTimingFd):
            return False

        return self._data == other._data

    def __hash__(self) -> int:
        return tuple(self._data.values()).__hash__()


def _oscillator_tolerance_condition_1(nom_sjw: int, nbt: int) -> float:
    """Arbitration phase - resynchronization"""
    return nom_sjw / (2 * 10 * nbt)


def _oscillator_tolerance_condition_2(
    nbt: int, nom_phase_seg1: int, nom_phase_seg2: int
) -> float:
    """Arbitration phase - sampling of bit after error flag"""
    return min(nom_phase_seg1, nom_phase_seg2) / (2 * (13 * nbt - nom_phase_seg2))


def _oscillator_tolerance_condition_3(data_sjw: int, dbt: int) -> float:
    """Data phase - resynchronization"""
    return data_sjw / (2 * 10 * dbt)


def _oscillator_tolerance_condition_4(
    nom_phase_seg1: int,
    nom_phase_seg2: int,
    data_phase_seg2: int,
    nbt: int,
    dbt: int,
    data_brp: int,
    nom_brp: int,
) -> float:
    """Data phase - sampling of bit after error flag"""
    return min(nom_phase_seg1, nom_phase_seg2) / (
        2 * ((6 * dbt - data_phase_seg2) * data_brp / nom_brp + 7 * nbt)
    )


def _oscillator_tolerance_condition_5(
    data_sjw: int,
    data_brp: int,
    nom_brp: int,
    nom_phase_seg2: int,
    data_phase_seg2: int,
    nbt: int,
    dbt: int,
) -> float:
    """Data phase - bit rate switch"""
    max_correctable_phase_shift = data_sjw - max(0.0, nom_brp / data_brp - 1)
    time_between_resync = 2 * (
        (2 * nbt - nom_phase_seg2) * nom_brp / data_brp + data_phase_seg2 + 4 * dbt
    )
    return max_correctable_phase_shift / time_between_resync



"""
The main module of the socketcan interface containing most user-facing classes and methods
along some internal methods.

At the end of the file the usage of the internal methods is shown.
"""

import ctypes
import ctypes.util
import errno
import logging
import select
import socket
import struct
import threading
import time
from typing import Dict, List, Optional, Sequence, Tuple, Type, Union

log = logging.getLogger(__name__)
log_tx = log.getChild("tx")
log_rx = log.getChild("rx")

try:
    from socket import CMSG_SPACE

    CMSG_SPACE_available = True
except ImportError:
    CMSG_SPACE_available = False
    log.error("socket.CMSG_SPACE not available on this platform")


import can
from can import BusABC, Message
from can.broadcastmanager import (
    LimitedDurationCyclicSendTaskABC,
    ModifiableCyclicTaskABC,
    RestartableCyclicTaskABC,
)
from can.interfaces.socketcan import constants
from can.interfaces.socketcan.utils import find_available_interfaces, pack_filters
from can.typechecking import CanFilters


# Setup BCM struct
def bcm_header_factory(
    fields: List[Tuple[str, Union[Type[ctypes.c_uint32], Type[ctypes.c_long]]]],
    alignment: int = 8,
):
    curr_stride = 0
    results: List[
        Tuple[
            str, Union[Type[ctypes.c_uint8], Type[ctypes.c_uint32], Type[ctypes.c_long]]
        ]
    ] = []
    pad_index = 0
    for field in fields:
        field_alignment = ctypes.alignment(field[1])
        field_size = ctypes.sizeof(field[1])

        # If the current stride index isn't a multiple of the alignment
        # requirements of this field, then we must add padding bytes until we
        # are aligned
        while curr_stride % field_alignment != 0:
            results.append((f"pad_{pad_index}", ctypes.c_uint8))
            pad_index += 1
            curr_stride += 1

        # Now can it fit?
        # Example: If this is 8 bytes and the type requires 4 bytes alignment
        # then we can only fit when we're starting at 0. Otherwise, we will
        # split across 2 strides.
        #
        # | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 |
        results.append(field)
        curr_stride += field_size

    # Add trailing padding to align to a multiple of the largest scalar member
    # in the structure
    while curr_stride % alignment != 0:
        results.append((f"pad_{pad_index}", ctypes.c_uint8))
        pad_index += 1
        curr_stride += 1

    return type("BcmMsgHead", (ctypes.Structure,), {"_fields_": results})


# The fields definition is taken from the C struct definitions in
# <linux/can/bcm.h>
#
#     struct bcm_timeval {
#     	long tv_sec;
#     	long tv_usec;
#     };
#
#     /**
#      * struct bcm_msg_head - head of messages to/from the broadcast manager
#      * @opcode:    opcode, see enum below.
#      * @flags:     special flags, see below.
#      * @count:     number of frames to send before changing interval.
#      * @ival1:     interval for the first @count frames.
#      * @ival2:     interval for the following frames.
#      * @can_id:    CAN ID of frames to be sent or received.
#      * @nframes:   number of frames appended to the message head.
#      * @frames:    array of CAN frames.
#      */
#     struct bcm_msg_head {
#     	__u32 opcode;
#     	__u32 flags;
#     	__u32 count;
#     	struct bcm_timeval ival1, ival2;
#     	canid_t can_id;
#     	__u32 nframes;
#     	struct can_frame frames[0];
#     };
BcmMsgHead = bcm_header_factory(
    fields=[
        ("opcode", ctypes.c_uint32),
        ("flags", ctypes.c_uint32),
        ("count", ctypes.c_uint32),
        ("ival1_tv_sec", ctypes.c_long),
        ("ival1_tv_usec", ctypes.c_long),
        ("ival2_tv_sec", ctypes.c_long),
        ("ival2_tv_usec", ctypes.c_long),
        ("can_id", ctypes.c_uint32),
        ("nframes", ctypes.c_uint32),
    ]
)


# struct module defines a binary packing format:
# https://docs.python.org/3/library/struct.html#struct-format-strings
# The 32bit can id is directly followed by the 8bit data link count
# The data field is aligned on an 8 byte boundary, hence we add padding
# which aligns the data field to an 8 byte boundary.
CAN_FRAME_HEADER_STRUCT = struct.Struct("=IBB2x")


def build_can_frame(msg: Message) -> bytes:
    """CAN frame packing/unpacking (see 'struct can_frame' in <linux/can.h>)
    /**
     * struct can_frame - basic CAN frame structure
     * @can_id:  the CAN ID of the frame and CAN_*_FLAG flags, see above.
     * @can_dlc: the data length field of the CAN frame
     * @data:    the CAN frame payload.
     */
    struct can_frame {
        canid_t can_id;  /* 32 bit CAN_ID + EFF/RTR/ERR flags */
        __u8    can_dlc; /* data length code: 0 .. 8 */
        __u8    data[8] __attribute__((aligned(8)));
    };

    /**
    * struct canfd_frame - CAN flexible data rate frame structure
    * @can_id: CAN ID of the frame and CAN_*_FLAG flags, see canid_t definition
    * @len:    frame payload length in byte (0 .. CANFD_MAX_DLEN)
    * @flags:  additional flags for CAN FD
    * @__res0: reserved / padding
    * @__res1: reserved / padding
    * @data:   CAN FD frame payload (up to CANFD_MAX_DLEN byte)
    */
    struct canfd_frame {
        canid_t can_id;  /* 32 bit CAN_ID + EFF/RTR/ERR flags */
        __u8    len;     /* frame payload length in byte */
        __u8    flags;   /* additional flags for CAN FD */
        __u8    __res0;  /* reserved / padding */
        __u8    __res1;  /* reserved / padding */
        __u8    data[CANFD_MAX_DLEN] __attribute__((aligned(8)));
    };
    """
    can_id = _compose_arbitration_id(msg)
    flags = 0
    if msg.bitrate_switch:
        flags |= constants.CANFD_BRS
    if msg.error_state_indicator:
        flags |= constants.CANFD_ESI
    max_len = 64 if msg.is_fd else 8
    data = bytes(msg.data).ljust(max_len, b"\x00")
    return CAN_FRAME_HEADER_STRUCT.pack(can_id, msg.dlc, flags) + data


def build_bcm_header(
    opcode: int,
    flags: int,
    count: int,
    ival1_seconds: int,
    ival1_usec: int,
    ival2_seconds: int,
    ival2_usec: int,
    can_id: int,
    nframes: int,
) -> bytes:
    result = BcmMsgHead(
        opcode=opcode,
        flags=flags,
        count=count,
        ival1_tv_sec=ival1_seconds,
        ival1_tv_usec=ival1_usec,
        ival2_tv_sec=ival2_seconds,
        ival2_tv_usec=ival2_usec,
        can_id=can_id,
        nframes=nframes,
    )
    return ctypes.string_at(ctypes.addressof(result), ctypes.sizeof(result))


def build_bcm_tx_delete_header(can_id: int, flags: int) -> bytes:
    opcode = constants.CAN_BCM_TX_DELETE
    return build_bcm_header(opcode, flags, 0, 0, 0, 0, 0, can_id, 1)


def build_bcm_transmit_header(
    can_id: int,
    count: int,
    initial_period: float,
    subsequent_period: float,
    msg_flags: int,
    nframes: int = 1,
) -> bytes:
    opcode = constants.CAN_BCM_TX_SETUP

    flags = msg_flags | constants.SETTIMER | constants.STARTTIMER

    if initial_period > 0:
        # Note `TX_COUNTEVT` creates the message TX_EXPIRED when count expires
        flags |= constants.TX_COUNTEVT

    def split_time(value: float) -> Tuple[int, int]:
        """Given seconds as a float, return whole seconds and microseconds"""
        seconds = int(value)
        microseconds = int(1e6 * (value - seconds))
        return seconds, microseconds

    ival1_seconds, ival1_usec = split_time(initial_period)
    ival2_seconds, ival2_usec = split_time(subsequent_period)

    return build_bcm_header(
        opcode,
        flags,
        count,
        ival1_seconds,
        ival1_usec,
        ival2_seconds,
        ival2_usec,
        can_id,
        nframes,
    )


def build_bcm_update_header(can_id: int, msg_flags: int, nframes: int = 1) -> bytes:
    return build_bcm_header(
        constants.CAN_BCM_TX_SETUP, msg_flags, 0, 0, 0, 0, 0, can_id, nframes
    )


def dissect_can_frame(frame: bytes) -> Tuple[int, int, int, bytes]:
    can_id, can_dlc, flags = CAN_FRAME_HEADER_STRUCT.unpack_from(frame)
    if len(frame) != constants.CANFD_MTU:
        # Flags not valid in non-FD frames
        flags = 0
    return can_id, can_dlc, flags, frame[8 : 8 + can_dlc]


def create_bcm_socket(channel: str) -> socket.socket:
    """create a broadcast manager socket and connect to the given interface"""
    s = socket.socket(constants.PF_CAN, socket.SOCK_DGRAM, constants.CAN_BCM)
    s.connect((channel,))
    return s


def send_bcm(bcm_socket: socket.socket, data: bytes) -> int:
    """
    Send raw frame to a BCM socket and handle errors.
    """
    try:
        return bcm_socket.send(data)
    except OSError as error:
        base = f"Couldn't send CAN BCM frame due to OS Error: {error.strerror}"

        if error.errno == errno.EINVAL:
            specific_message = " You are probably referring to a non-existing frame."
        elif error.errno == errno.ENETDOWN:
            specific_message = " The CAN interface appears to be down."
        elif error.errno == errno.EBADF:
            specific_message = " The CAN socket appears to be closed."
        else:
            specific_message = ""

        raise can.CanOperationError(base + specific_message, error.errno) from error


def _compose_arbitration_id(message: Message) -> int:
    can_id = message.arbitration_id
    if message.is_extended_id:
        log.debug("sending an extended id type message")
        can_id |= constants.CAN_EFF_FLAG
    if message.is_remote_frame:
        log.debug("requesting a remote frame")
        can_id |= constants.CAN_RTR_FLAG
    if message.is_error_frame:
        log.debug("sending error frame")
        can_id |= constants.CAN_ERR_FLAG
    return can_id


class CyclicSendTask(
    LimitedDurationCyclicSendTaskABC, ModifiableCyclicTaskABC, RestartableCyclicTaskABC
):
    """
    A SocketCAN cyclic send task supports:

        - setting of a task duration
        - modifying the data
        - stopping then subsequent restarting of the task
    """

    def __init__(
        self,
        bcm_socket: socket.socket,
        task_id: int,
        messages: Union[Sequence[Message], Message],
        period: float,
        duration: Optional[float] = None,
    ) -> None:
        """Construct and :meth:`~start` a task.

        :param bcm_socket: An open BCM socket on the desired CAN channel.
        :param task_id:
            The identifier used to uniquely reference particular cyclic send task
            within Linux BCM.
        :param messages:
            The messages to be sent periodically.
        :param period:
            The rate in seconds at which to send the messages.
        :param duration:
            Approximate duration in seconds to send the messages for.
        """
        # The following are assigned by LimitedDurationCyclicSendTaskABC:
        #   - self.messages
        #   - self.period
        #   - self.duration
        super().__init__(messages, period, duration)

        self.bcm_socket = bcm_socket
        self.task_id = task_id
        self._tx_setup(self.messages)

    def _tx_setup(
        self, messages: Sequence[Message], raise_if_task_exists: bool = True
    ) -> None:
        # Create a low level packed frame to pass to the kernel
        body = bytearray()
        self.flags = constants.CAN_FD_FRAME if messages[0].is_fd else 0

        if self.duration:
            count = int(self.duration / self.period)
            ival1 = self.period
            ival2 = 0.0
        else:
            count = 0
            ival1 = 0.0
            ival2 = self.period

        if raise_if_task_exists:
            self._check_bcm_task()

        header = build_bcm_transmit_header(
            self.task_id, count, ival1, ival2, self.flags, nframes=len(messages)
        )
        for message in messages:
            body += build_can_frame(message)
        log.debug("Sending BCM command")
        send_bcm(self.bcm_socket, header + body)

    def _check_bcm_task(self) -> None:
        # Do a TX_READ on a task ID, and check if we get EINVAL. If so,
        # then we are referring to a CAN message with an existing ID
        check_header = build_bcm_header(
            opcode=constants.CAN_BCM_TX_READ,
            flags=0,
            count=0,
            ival1_seconds=0,
            ival1_usec=0,
            ival2_seconds=0,
            ival2_usec=0,
            can_id=self.task_id,
            nframes=0,
        )
        log.debug(
            "Reading properties of (cyclic) transmission task id=%d", self.task_id
        )
        try:
            self.bcm_socket.send(check_header)
        except OSError as error:
            if error.errno != errno.EINVAL:
                raise can.CanOperationError("failed to check", error.errno) from error
            else:
                log.debug("Invalid argument - transmission task not known to kernel")
        else:
            # No exception raised - transmission task with this ID exists in kernel.
            # Existence of an existing transmission task might not be a problem!
            raise can.CanOperationError(
                f"A periodic task for task ID {self.task_id} is already in progress "
                "by the SocketCAN Linux layer"
            )

    def stop(self) -> None:
        """Stop a task by sending TX_DELETE message to Linux kernel.

        This will delete the entry for the transmission of the CAN-message
        with the specified ``task_id`` identifier. The message length
        for the command TX_DELETE is {[bcm_msg_head]} (only the header).
        """
        log.debug("Stopping periodic task")

        stopframe = build_bcm_tx_delete_header(self.task_id, self.flags)
        send_bcm(self.bcm_socket, stopframe)

    def modify_data(self, messages: Union[Sequence[Message], Message]) -> None:
        """Update the contents of the periodically sent CAN messages by
        sending TX_SETUP message to Linux kernel.

        The number of new cyclic messages to be sent must be equal to the
        original number of messages originally specified for this task.

        .. note:: The messages must all have the same
                  :attr:`~can.Message.arbitration_id` like the first message.

        :param messages:
            The messages with the new :attr:`can.Message.data`.
        """
        messages = self._check_and_convert_messages(messages)
        self._check_modified_messages(messages)

        self.messages = messages

        body = bytearray()
        header = build_bcm_update_header(
            can_id=self.task_id, msg_flags=self.flags, nframes=len(messages)
        )
        for message in messages:
            body += build_can_frame(message)
        log.debug("Sending BCM command")
        send_bcm(self.bcm_socket, header + body)

    def start(self) -> None:
        """Restart a periodic task by sending TX_SETUP message to Linux kernel.

        It verifies presence of the particular BCM task through sending TX_READ
        message to Linux kernel prior to scheduling.

        :raises ValueError:
            If the task referenced by ``task_id`` is already running.
        """
        self._tx_setup(self.messages, raise_if_task_exists=False)


class MultiRateCyclicSendTask(CyclicSendTask):
    """Exposes more of the full power of the TX_SETUP opcode."""

    def __init__(
        self,
        channel: socket.socket,
        task_id: int,
        messages: Sequence[Message],
        count: int,
        initial_period: float,
        subsequent_period: float,
    ):
        super().__init__(channel, task_id, messages, subsequent_period)

        # Create a low level packed frame to pass to the kernel
        header = build_bcm_transmit_header(
            self.task_id,
            count,
            initial_period,
            subsequent_period,
            self.flags,
            nframes=len(messages),
        )

        body = bytearray()
        for message in messages:
            body += build_can_frame(message)

        log.info("Sending BCM TX_SETUP command")
        send_bcm(self.bcm_socket, header + body)


def create_socket() -> socket.socket:
    """Creates a raw CAN socket. The socket will
    be returned unbound to any interface.
    """
    sock = socket.socket(constants.PF_CAN, socket.SOCK_RAW, constants.CAN_RAW)

    log.info("Created a socket")

    return sock


def bind_socket(sock: socket.socket, channel: str = "can0") -> None:
    """
    Binds the given socket to the given interface.

    :param sock:
        The socket to be bound
    :param channel:
        The channel / interface to bind to
    :raises OSError:
        If the specified interface isn't found.
    """
    log.debug("Binding socket to channel=%s", channel)
    sock.bind((channel,))
    log.debug("Bound socket.")


def capture_message(
    sock: socket.socket, get_channel: bool = False
) -> Optional[Message]:
    """
    Captures a message from given socket.

    :param sock:
        The socket to read a message from.
    :param get_channel:
        Find out which channel the message comes from.

    :return: The received message, or None on failure.
    """
    # Fetching the Arb ID, DLC and Data
    try:
        cf, ancillary_data, msg_flags, addr = sock.recvmsg(
            constants.CANFD_MTU, RECEIVED_ANCILLARY_BUFFER_SIZE
        )
        if get_channel:
            channel = addr[0] if isinstance(addr, tuple) else addr
        else:
            channel = None
    except OSError as error:
        raise can.CanOperationError(f"Error receiving: {error.strerror}", error.errno)

    can_id, can_dlc, flags, data = dissect_can_frame(cf)

    # Fetching the timestamp
    assert len(ancillary_data) == 1, "only requested a single extra field"
    cmsg_level, cmsg_type, cmsg_data = ancillary_data[0]
    assert (
        cmsg_level == socket.SOL_SOCKET and cmsg_type == constants.SO_TIMESTAMPNS
    ), "received control message type that was not requested"
    # see https://man7.org/linux/man-pages/man3/timespec.3.html -> struct timespec for details
    seconds, nanoseconds = RECEIVED_TIMESTAMP_STRUCT.unpack_from(cmsg_data)
    if nanoseconds >= 1e9:
        raise can.CanOperationError(
            f"Timestamp nanoseconds field was out of range: {nanoseconds} not less than 1e9"
        )
    timestamp = seconds + nanoseconds * 1e-9

    # EXT, RTR, ERR flags -> boolean attributes
    #   /* special address description flags for the CAN_ID */
    #   #define CAN_EFF_FLAG 0x80000000U /* EFF/SFF is set in the MSB */
    #   #define CAN_RTR_FLAG 0x40000000U /* remote transmission request */
    #   #define CAN_ERR_FLAG 0x20000000U /* error frame */
    is_extended_frame_format = bool(can_id & constants.CAN_EFF_FLAG)
    is_remote_transmission_request = bool(can_id & constants.CAN_RTR_FLAG)
    is_error_frame = bool(can_id & constants.CAN_ERR_FLAG)
    is_fd = len(cf) == constants.CANFD_MTU
    bitrate_switch = bool(flags & constants.CANFD_BRS)
    error_state_indicator = bool(flags & constants.CANFD_ESI)

    # Section 4.7.1: MSG_DONTROUTE: set when the received frame was created on the local host.
    is_rx = not bool(msg_flags & socket.MSG_DONTROUTE)

    if is_extended_frame_format:
        # log.debug("CAN: Extended")
        # TODO does this depend on SFF or EFF?
        arbitration_id = can_id & 0x1FFFFFFF
    else:
        # log.debug("CAN: Standard")
        arbitration_id = can_id & 0x000007FF

    msg = Message(
        timestamp=timestamp,
        channel=channel,
        arbitration_id=arbitration_id,
        is_extended_id=is_extended_frame_format,
        is_remote_frame=is_remote_transmission_request,
        is_error_frame=is_error_frame,
        is_fd=is_fd,
        is_rx=is_rx,
        bitrate_switch=bitrate_switch,
        error_state_indicator=error_state_indicator,
        dlc=can_dlc,
        data=data,
    )

    return msg


# Constants needed for precise handling of timestamps
if CMSG_SPACE_available:
    RECEIVED_TIMESTAMP_STRUCT = struct.Struct("@ll")
    RECEIVED_ANCILLARY_BUFFER_SIZE = CMSG_SPACE(RECEIVED_TIMESTAMP_STRUCT.size)


class SocketcanBus(BusABC):
    """A SocketCAN interface to CAN.

    It implements :meth:`can.BusABC._detect_available_configs` to search for
    available interfaces.
    """

    def __init__(
        self,
        channel: str = "",
        receive_own_messages: bool = False,
        local_loopback: bool = True,
        fd: bool = False,
        can_filters: Optional[CanFilters] = None,
        ignore_rx_error_frames=False,
        **kwargs,
    ) -> None:
        """Creates a new socketcan bus.

        If setting some socket options fails, an error will be printed
        but no exception will be thrown. This includes enabling:

            - that own messages should be received,
            - CAN-FD frames and
            - error frames.

        :param channel:
            The can interface name with which to create this bus.
            An example channel would be 'vcan0' or 'can0'.
            An empty string '' will receive messages from all channels.
            In that case any sent messages must be explicitly addressed to a
            channel using :attr:`can.Message.channel`.
        :param receive_own_messages:
            If transmitted messages should also be received by this bus.
        :param local_loopback:
            If local loopback should be enabled on this bus.
            Please note that local loopback does not mean that messages sent
            on a socket will be readable on the same socket, they will only
            be readable on other open sockets on the same machine. More info
            can be read on the socketcan documentation:
            See https://www.kernel.org/doc/html/latest/networking/can.html#socketcan-local-loopback1
        :param fd:
            If CAN-FD frames should be supported.
        :param can_filters:
            See :meth:`can.BusABC.set_filters`.
        :param ignore_rx_error_frames:
            If incoming error frames should be discarded.
        """
        self.socket = create_socket()
        self.channel = channel
        self.channel_info = f"socketcan channel '{channel}'"
        self._bcm_sockets: Dict[str, socket.socket] = {}
        self._is_filtered = False
        self._task_id = 0
        self._task_id_guard = threading.Lock()

        # set the local_loopback parameter
        try:
            self.socket.setsockopt(
                constants.SOL_CAN_RAW,
                constants.CAN_RAW_LOOPBACK,
                1 if local_loopback else 0,
            )
        except OSError as error:
            log.error("Could not set local loopback flag(%s)", error)

        # set the receive_own_messages parameter
        try:
            self.socket.setsockopt(
                constants.SOL_CAN_RAW,
                constants.CAN_RAW_RECV_OWN_MSGS,
                1 if receive_own_messages else 0,
            )
        except OSError as error:
            log.error("Could not receive own messages (%s)", error)

        # enable CAN-FD frames if desired
        if fd:
            try:
                self.socket.setsockopt(
                    constants.SOL_CAN_RAW, constants.CAN_RAW_FD_FRAMES, 1
                )
            except OSError as error:
                log.error("Could not enable CAN-FD frames (%s)", error)

        if not ignore_rx_error_frames:
            # enable error frames
            try:
                self.socket.setsockopt(
                    constants.SOL_CAN_RAW, constants.CAN_RAW_ERR_FILTER, 0x1FFFFFFF
                )
            except OSError as error:
                log.error("Could not enable error frames (%s)", error)

        # enable nanosecond resolution timestamping
        # we can always do this since
        #  1) it is guaranteed to be at least as precise as without
        #  2) it is available since Linux 2.6.22, and CAN support was only added afterward
        #     so this is always supported by the kernel
        self.socket.setsockopt(socket.SOL_SOCKET, constants.SO_TIMESTAMPNS, 1)

        bind_socket(self.socket, channel)
        kwargs.update(
            {
                "receive_own_messages": receive_own_messages,
                "fd": fd,
                "local_loopback": local_loopback,
            }
        )
        super().__init__(channel=channel, can_filters=can_filters, **kwargs)

    def shutdown(self) -> None:
        """Stops all active periodic tasks and closes the socket."""
        super().shutdown()
        for channel, bcm_socket in self._bcm_sockets.items():
            log.debug("Closing bcm socket for channel %s", channel)
            bcm_socket.close()
        log.debug("Closing raw can socket")
        self.socket.close()

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> Tuple[Optional[Message], bool]:
        try:
            # get all sockets that are ready (can be a list with a single value
            # being self.socket or an empty list if self.socket is not ready)
            ready_receive_sockets, _, _ = select.select([self.socket], [], [], timeout)
        except OSError as error:
            # something bad happened (e.g. the interface went down)
            raise can.CanOperationError(
                f"Failed to receive: {error.strerror}", error.errno
            )

        if ready_receive_sockets:  # not empty
            get_channel = self.channel == ""
            msg = capture_message(self.socket, get_channel)
            if msg and not msg.channel and self.channel:
                # Default to our own channel
                msg.channel = self.channel
            return msg, self._is_filtered

        # socket wasn't readable or timeout occurred
        return None, self._is_filtered

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        """Transmit a message to the CAN bus.

        :param msg: A message object.
        :param timeout:
            Wait up to this many seconds for the transmit queue to be ready.
            If not given, the call may fail immediately.

        :raises ~can.exceptions.CanError:
            if the message could not be written.
        """
        log.debug("We've been asked to write a message to the bus")
        logger_tx = log.getChild("tx")
        logger_tx.debug("sending: %s", msg)

        started = time.time()
        # If no timeout is given, poll for availability
        if timeout is None:
            timeout = 0
        time_left = timeout
        data = build_can_frame(msg)

        while time_left >= 0:
            # Wait for write availability
            ready = select.select([], [self.socket], [], time_left)[1]
            if not ready:
                # Timeout
                break
            channel = str(msg.channel) if msg.channel else None
            sent = self._send_once(data, channel)
            if sent == len(data):
                return
            # Not all data were sent, try again with remaining data
            data = data[sent:]
            time_left = timeout - (time.time() - started)

        raise can.CanOperationError("Transmit buffer full")

    def _send_once(self, data: bytes, channel: Optional[str] = None) -> int:
        try:
            if self.channel == "" and channel:
                # Message must be addressed to a specific channel
                sent = self.socket.sendto(data, (channel,))
            else:
                sent = self.socket.send(data)
        except OSError as error:
            raise can.CanOperationError(
                f"Failed to transmit: {error.strerror}", error.errno
            )
        return sent

    def _send_periodic_internal(
        self,
        msgs: Union[Sequence[Message], Message],
        period: float,
        duration: Optional[float] = None,
    ) -> CyclicSendTask:
        """Start sending messages at a given period on this bus.

        The Linux kernel's Broadcast Manager SocketCAN API is used to schedule
        periodic sending of CAN messages. The wrapping 32-bit counter (see
        :meth:`~_get_next_task_id()`) designated to distinguish different
        :class:`CyclicSendTask` within BCM provides flexibility to schedule
        CAN messages sending with the same CAN ID, but different CAN data.

        :param messages:
            The message(s) to be sent periodically.
        :param period:
            The rate in seconds at which to send the messages.
        :param duration:
            Approximate duration in seconds to continue sending messages. If
            no duration is provided, the task will continue indefinitely.

        :raises ValueError:
            If task identifier passed to :class:`CyclicSendTask` can't be used
            to schedule new task in Linux BCM.

        :return:
            A :class:`CyclicSendTask` task instance. This can be used to modify the data,
            pause/resume the transmission and to stop the transmission.

        .. note::

            Note the duration before the messages stop being sent may not
            be exactly the same as the duration specified by the user. In
            general the message will be sent at the given rate until at
            least *duration* seconds.
        """
        msgs = LimitedDurationCyclicSendTaskABC._check_and_convert_messages(  # pylint: disable=protected-access
            msgs
        )

        msgs_channel = str(msgs[0].channel) if msgs[0].channel else None
        bcm_socket = self._get_bcm_socket(msgs_channel or self.channel)
        task_id = self._get_next_task_id()
        task = CyclicSendTask(bcm_socket, task_id, msgs, period, duration)
        return task

    def _get_next_task_id(self) -> int:
        with self._task_id_guard:
            self._task_id = (self._task_id + 1) % (2**32 - 1)
            return self._task_id

    def _get_bcm_socket(self, channel: str) -> socket.socket:
        if channel not in self._bcm_sockets:
            self._bcm_sockets[channel] = create_bcm_socket(self.channel)
        return self._bcm_sockets[channel]

    def _apply_filters(self, filters: Optional[can.typechecking.CanFilters]) -> None:
        try:
            self.socket.setsockopt(
                constants.SOL_CAN_RAW, constants.CAN_RAW_FILTER, pack_filters(filters)
            )
        except OSError as error:
            # fall back to "software filtering" (= not in kernel)
            self._is_filtered = False
            log.error(
                "Setting filters failed; falling back to software filtering (not in kernel): %s",
                error,
            )
        else:
            self._is_filtered = True

    def fileno(self) -> int:
        return self.socket.fileno()

    @staticmethod
    def _detect_available_configs() -> List[can.typechecking.AutoDetectedConfig]:
        return [
            {"interface": "socketcan", "channel": channel}
            for channel in find_available_interfaces()
        ]


if __name__ == "__main__":
    # This example demonstrates how to use the internal methods of this module.
    # It creates two sockets on vcan0 to test sending and receiving.
    #
    # If you want to try it out you can do the following (possibly using sudo):
    #
    #     modprobe vcan
    #     ip link add dev vcan0 type vcan
    #     ip link set vcan0 up
    #
    log.setLevel(logging.DEBUG)

    def receiver(event: threading.Event) -> None:
        receiver_socket = create_socket()
        bind_socket(receiver_socket, "vcan0")
        print("Receiver is waiting for a message...")
        event.set()
        print(f"Receiver got: {capture_message(receiver_socket)}")

    def sender(event: threading.Event) -> None:
        event.wait()
        sender_socket = create_socket()
        bind_socket(sender_socket, "vcan0")
        msg = Message(arbitration_id=0x01, data=b"\x01\x02\x03")
        sender_socket.send(build_can_frame(msg))
        print("Sender sent a message.")

    e = threading.Event()
    threading.Thread(target=receiver, args=(e,)).start()
    threading.Thread(target=sender, args=(e,)).start()
